/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitAddPdf.cc,v 1.5 2005/07/03 23:04:22 bevan Exp $
 * Authors:
 *    A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFit] --
//
//  AFitAddPdf(RooAbsReal& x, TString name, TString types) :
//
// This class builds a composite PDF (RooAddPdf) based in a comma separated 
// list of input types for a dependant variable x.  The name for
// the PDF is given by name, all component PDFs have a name given by
// name_NN, where NN is an integer 0, 1, 2, ...  The string types,
// is the comma separated list of types.
//
// The RooAddPdf built by this class is of the form
//
//        c_1*PDF_1 + c_2*PDF_2 + ... (1-sum(c_1...c_n-1))*PDF_n
// 
// Example:
//    Make a double Gaussian + Argus function PDF
//
//       RooRealVar x("x", "", 0, 3);
//       AFitAbsPdfBuilder::setDataCard("datacardAddPdf.txt");
//       AFitAddPdf pdf(x, "doubleGaussian", "gaussian,gaussian,argus");
//       RooAbsPdf * PDF = pdf.getPdf();
//
//    where the datacard looks like
//
//       [doubleGaussian]
//       doubleGaussian_frac0 = 0.3 +/- 1.0
//       doubleGaussian_frac1 = 0.3 +/- 1.0
//       
//       [doubleGaussian_0]
//       doubleGaussian_0mean = 1.0 +/- 0.0
//       doubleGaussian_0width = 1.0 +/- 0.0
//       
//       [doubleGaussian_1]
//       doubleGaussian_1mean = 0.0 +/- 0.0
//       doubleGaussian_1width = 0.2 +/- 0.0
//       
//       [doubleGaussian_2]
//       doubleGaussian_2endpt = 3.0 +/- 0.0
//       doubleGaussian_2xi = -20 +/- 0.0
//       doubleGaussian_2power = 0.5 +/- 0.0
//
// Computer resources dictate the limit of the number of composite PDFs that
// can be constructed.  For example, multi Gaussian PDFs can be implemented 
// using this class.
//

#include "AFit/AFitAddPdf.hh"
#ifdef __BABARROOFIT__
#include "RooFitCore/RooAddPdf.hh"
#else
#include "RooAddPdf.h"
#endif

#include "TString.h"
#include "TObjString.h"
#include "TObjArray.h"

ClassImp(AFitAddPdf)
  ;

AFitAddPdf::AFitAddPdf(RooAbsReal& x, TString name, TString types) :
  _types("types", "Comma separated list of PDF types to build", types)
  , _configFile("AFitAddPdf") 
 {
  // Constructor
  _name  = name;
  _x     = &x;

  varSet.add(_types);

  readDataCard();
}

AFitAddPdf::~AFitAddPdf() {
  cout << ClassName() << ":destructor to implement DELETE THE INSTANTIATED FRACS AND PDFS!" << endl;
}

void AFitAddPdf::buildPdf() {
  if(condVarSet) fact.setConditionalVariables(*condVarSet);
  fact.setConfigFile(_configFile);

  // Build the Pdf
  TString typestring = _types;

  TObjArray * typArr = typestring.Tokenize(',');
  Int_t nTypes = typArr->GetEntries();

  for(int i = 0; i < nTypes; i++) {
    TString thisType = ((TObjString*)typArr->At(i))->GetString();
    TString thisName = _name; thisName += "_"; thisName += (long)i; 
    AFitAbsPdfBuilder * bld = (AFitAbsPdfBuilder *)fact.makePdf(thisName, thisType, *_x);
    bldList.Add(bld);
    pdfList.add(*bld->getPdf());
    if(i != nTypes -1) {
      TString fracName = _name; fracName += "_frac"; fracName += (long)i; 
      fracList.add(* new RooRealVar (fracName, "Composite PDF fraction", 0.5, 0.0, 1.0) );
    }
  }

  varSet.add(fracList);
  // re-read the data card to pick up fractions (if set)
  readDataCard();

  _thePdf = new RooAddPdf(_name, "Composite PDF from AFitAddPdf", pdfList, fracList);
}

void AFitAddPdf::writeDataCard(ostream& out) {
  if(!writtenToConfigFile) {
    out << "["<< _name << "]" << endl;
    varSet.writeToStream(out, false);
    out << endl << endl;

    // now iterate over the pdf list of built PDFs & write these to the configuration file
    TIterator * iter = bldList.MakeIterator();
    TIter next(iter);
    AFitAbsPdfBuilder * obj=0;
    while (obj = (AFitAbsPdfBuilder *)next()) {
      out << "// - this is an AFitAbsPdfBuilder " << obj->_name << " ( from " << ClassName() << ")" << endl;
      obj->writeDataCard(out);
      out << endl<< endl;
    }

    writtenToConfigFile = kTRUE;
  }
}
