/*****************************************************************************
 * Project: RooFit                                                           *
 * Package: RooFitModels                                                     *
 *    File: $Id: AFitFlattePdf.cc,v 1.1 2007/02/13 09:35:44 bevan Exp $
 * Authors:                                                                  *
 * Katherine George - Queen Mary University of London                        *
 * Adrian Bevan (ported from RptBaseTools to AFit - 2009)                    *
 *****************************************************************************/

// -- CLASS DESCRIPTION [PDF] --


#include <iostream>
#include <math.h>

#include "AFit/AFitFlattePdf.hh"

#ifdef __BABARROOFIT__
// RooFit includes
#include "RooFitCore/RooAbsReal.hh"
#include "RooFitCore/RooRealVar.hh"
#else
// RooFit includes
#include "RooAbsReal.h"
#include "RooRealVar.h"
#endif


using namespace std;

ClassImp(AFitFlattePdf)

AFitFlattePdf::AFitFlattePdf(const char *name, const char *title,
			 RooAbsReal& _x, RooAbsReal& _mean,
			 RooAbsReal& _gpi,RooAbsReal& _gk, RooAbsReal & _m0) :
  RooAbsPdf(name,title),
  x("x","Dependent",this,_x),
  mean("mean","Mass of resonance",this,_mean),
  gpi("gpi","coupling constant g_{pi}",this,_gpi),
  gk("gk","coupling constant g_{K}",this,_gk),
  m0("m0", "threshold mass", this, _m0) 
{
  /*
   * x    = m_hh mass calculated (dependent variable)
   * mean = resonance mass mean value
   * gpi  = constant related to width of pipi final state
   * gk   = constant related to width of KK final state
   * m0   = threshold mass (e.g. kaon mass)
   */
}


AFitFlattePdf::AFitFlattePdf(const AFitFlattePdf& other, const char* name) : 
  RooAbsPdf(other,name), x("x",this,other.x), mean("mean",this,other.mean),
  gpi("gpi", this, other.gpi), gk("gk", this, other.gk), m0("m0", this, other.m0)
{
}


Double_t AFitFlattePdf::evaluate() const
{
  // Use the definition from phipi BAD: 1268  for PDF = |A^2|, where
  // A is a complex amplitude.
  Double_t mr2 = mean*mean;
  Double_t x2 = x*x;
  Double_t m02 = m0*m0;
  Double_t dm = mr2-x2;
  Double_t fact_kk(0.0);
  Double_t Gamma_pipi = gpi*gpi;
  Double_t Gamma_kk   = gk*sqrt(fabs(x2/4.0 - m02));
  if(x > 2.0*m0) { // above production threshold for KK
    fact_kk = mr2*(Gamma_kk+Gamma_pipi)*(Gamma_kk+Gamma_pipi);
  } else { // below production threshold for KK
    //    Gamma_kk = gk*sqrt(m02/4.0 - x2); // imaginary width below threshold
    fact_kk = mr2*(Gamma_kk*Gamma_kk + Gamma_pipi+Gamma_pipi);
  }

  Double_t pdf(0.0);
  if((dm*dm + fact_kk) != 0.0) {
    pdf = mr2*Gamma_kk*Gamma_pipi / (dm*dm + fact_kk);
  }

  /*
   * calculate the phase space supression factor (leaving off constant normalization term of a half)
   * to avoid problems with self normalization of the PDF.
   */
  if(x2 - 4*m02 >0)  pdf *= sqrt(x2 - 4*m02)  ;
  else pdf = 0;

  //  cout << x << " " << pdf << " " << dm << " " << fact_kk << " " << dm*dm + fact_kk << " " <<  Gamma_kk << endl;
  return pdf;
}


/*
  Int_t AFitFlattePdf::getAnalyticalIntegral(RooArgSet& allVars, RooArgSet& analVars, const char* rangeName) const 
  {
  if (matchArgs(allVars,analVars,x)) return 1 ;
  return 0 ;
  }
  
  
  Double_t AFitFlattePdf::analyticalIntegral(Int_t code, const char* rangeName) const 
  {
  switch(code) {
  case 1: 
  {
  Double_t c = 2./width;
  return c*(atan(c*(x.max(rangeName)-mean)) - atan(c*(x.min(rangeName)-mean)));
  }
  }
  
  assert(0) ;
  return 0 ;
  }
*/

