/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitProjectionPlot.cc,v 1.5 2005/07/03 23:04:22 bevan Exp $
 * Authors:
 *    A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2007 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFit] --
// Utility interface to make projection plots of various types

#include "AFit/AFitProjectionPlot.hh"
#ifdef __BABARROOFIT__
#include "RooFitCore/RooPlot.hh"
#include "RooFitCore/RooGlobalFunc.hh"
#include "RooFitCore/RooRealVar.hh"
#else
#include "RooPlot.h"
#include "RooGlobalFunc.h"
#include "RooRealVar.h"
#endif

#include "TIterator.h"
#include "TAxis.h"

using namespace RooFit;

ClassImp(AFitProjectionPlot)
  ;

AFitProjectionPlot::AFitProjectionPlot() 
  : _scale(1.0)
{
}

RooPlot *   AFitProjectionPlot::makeAsymmetryPlot(RooRealVar & var,  RooCategory & asymVar, RooDataSet * data,  RooAbsPdf * pdf, RooDataSet * proto) {
  //
  // fill a RooPlot with data and a projection of the asymmetry, if necessary use
  // a prototype data set when plotting the PDF.
  //
  cout << "\tplotting asymmetry for " << var.GetName() << " with asymmetry category " << asymVar.GetName() << endl;

  RooPlot * frame = var.frame();
  if(data) data->plotOn(frame, RooFit::Asymmetry(asymVar) );

  if(pdf) {
    if(proto)
    {
      cout << "using a prototype to plot variable" << endl;
      pdf->plotOn(frame, RooFit::ProjWData(*proto), RooFit::Asymmetry(asymVar) );
    }
    else    {
      cout << "plotting variable from PDF - NOT using a prototype" << endl;
      pdf->plotOn(frame, RooFit::Asymmetry(asymVar) );
    }
  }
  setFrameTitles(frame, "", var.GetTitle(), "Asymmetry", _scale);

  return frame;
}

RooPlot * AFitProjectionPlot::makePlot(RooRealVar & var, RooDataSet * data,  RooAbsPdf * pdf, RooDataSet * proto, RooAbsPdf * newpdf)
{
  //
  // fill a RooPlot with data and a projection of the PDF - use the 
  // data set parsed to make a prototype data set and project the pdf
  // It is possible to also plot a second PDF and to use a prototype data
  // set when making the projection.  
  //
  // The plotted function will be a 1 dimensional integral of all of the data. 
  cout << "\tplotting " << var.GetName() << endl;
  cout << "\tRange: " << var.getMin() << " to " << var.getMax() << endl;


  RooPlot * frame = var.frame();
  if(data) data->plotOn(frame);
  if(pdf) {
    if(proto)
    {
      cout << "using a prototype to plot variable" << endl;
      pdf->plotOn(frame, ProjWData(*proto) );
      if(newpdf) 	{
	newpdf->plotOn(frame, ProjWData(*proto), LineColor(kRed), LineStyle(kDashed) );
      }
    }
    else    {
      cout << "plotting variable from PDF - NOT using a prototype" << endl;
      pdf->plotOn(frame);
      if(newpdf) 	{
	newpdf->plotOn(frame);
	frame->getAttLine()->SetLineColor(kRed);
	frame->getAttLine()->SetLineStyle(kDashed);
      }
    }
  }
  setFrameTitles(frame, "", var.GetTitle(), "Entries", _scale);

  return frame;
}

RooPlot * AFitProjectionPlot::makePlot(TString cut, RooRealVar & var, RooDataSet * data,  RooAbsPdf * pdf, RooDataSet * proto)
{
  //
  // fill a RooPlot with data and a projection of the PDF - use the 
  // data set parsed to make a prototype data set and project the pdf
  // If a cut string is supplied, then the data will be reduced before making the projection
  //
  // The plotted function will be a 1 dimensional integral of all of the data. 
  //
  // If a cut is to be applied to the data when plotting, in order to correct project over the prototype data,
  // this cut should also be applied to the proto dataset passed to this function.  The data set 'data' will
  // have the cut applied to it.  The reason for the different treatment of proto and data is that data
  // should have the variables cut upon in order to project onto the frame, however proto will not
  // necessarity have those same variables.
  RooPlot * frame = var.frame();
  RooDataSet * reducedData(0);
  if(data) {
    reducedData = (RooDataSet *)data->reduce(cut);
    reducedData->plotOn(frame, Cut(cut));
  }
  if(pdf) {
    if(proto) {      pdf->plotOn(frame,  Cut(cut), ProjWData(*proto));    }
    else pdf->plotOn(frame,  Cut(cut), ProjWData(*reducedData));
  }
  setFrameTitles(frame, "", var.GetTitle(), "Entries", _scale);

  return frame;
}


RooHist* AFitProjectionPlot::makePullPlot(RooHist* dhist, RooCurve* curve)
{
  // 'Borrowed' from the ChiSquare routine
  //
  // Given the RooHist from a RooPlot and the RooCurve from that plot, compute the pull 
  // distribution between the curve and data for all of the bins.
  //
  // This function was taken from PlotPull.cxx in the BaBar package S2bRFScripts/cxx/
  // 
  if(!dhist) {
    cout << "AFitProjectionPlot:ERROR in makePullPlot(...) histogram is null" << endl;
    return 0;
  }
  if(!curve) {
    cout << "AFitProjectionPlot:ERROR in makePullPlot(...) curve is null" << endl;
    return 0;
  }
  RooHist* histres = new RooHist(*dhist); // copy the incoming one

  Int_t np = dhist->GetN() ;
  Int_t i=0;

  Double_t x,y,eyl,eyh ;
  Double_t hbinw2 = dhist->getNominalBinWidth()/2 ;

  Double_t chisq(0) ;
  for (i=0 ; i<np ; i++)     {    
    // Retrieve histogram contents
    dhist->GetPoint(i,x,y) ;
    Double_t eyl = dhist->GetEYlow()[i];
    Double_t eyh = dhist->GetEYhigh()[i];
    // Integrate function over this bin
    Double_t avg = curve->average(x-hbinw2,x+hbinw2) ;

    // if (y!=0) 
    {      
      Double_t pull = (y>avg) ? ((y-avg)/eyl) : ((y-avg)/eyh) ;
      histres->SetPoint(i,x,pull);
      histres->SetPointError(i,0,0,1.,1.);
    }
  }
  histres->SetTitle("Pulls between the data and projected curve");
  
  return(histres);
}

void AFitProjectionPlot::setFrameTitles(RooPlot * frame, TString name, TString xTitle, TString yTitle, double scale)
{
  /*
   * Set up common format RooPlot style and axis labels
   *
   * name is the name of the frame (to replace the default address name)
   * xTitle and yTitle are the absicca and ordinate axis lables
   * scale is the scale factor to apply to the label/title font sizes.
   */
  frame->SetName(name);
  frame->SetTitle("");
  frame->SetXTitle(xTitle);
  frame->SetYTitle(yTitle);

  Int_t nDivX                = 205;  // number of divisions in x and y  
  Int_t nDivY                = 205;

  Double_t lineWidth         = 2.0;       // global parameters
  Double_t labelSize         = 0.05;       // label and title size
  Double_t labelOffset       = 0.005;
  Double_t titleOffsetX      = 0.8;
  Double_t titleOffsetY      = 0.8;

  frame->SetNdivisions(nDivX, "X");
  frame->SetLabelSize(labelSize*scale, "X");
  frame->SetLabelOffset(labelOffset, "X");
  frame->SetTitleOffset(titleOffsetX*scale, "X");
  frame->SetTitleSize(labelSize*scale, "X");

  frame->SetNdivisions(nDivY, "Y");
  frame->SetLabelSize(labelSize*scale, "Y");
  frame->SetLabelOffset(labelOffset, "Y");
  frame->SetTitleOffset(titleOffsetY*scale, "Y");
  frame->SetTitleSize(labelSize*scale, "Y");

  TAxis * ax = frame->GetXaxis();
  ax->CenterTitle();
  ax->SetTitleColor(kBlack);
  ax->SetTitleFont(132);
  ax->SetLabelFont(132);

  TAxis * ay = frame->GetYaxis();
  ay->CenterTitle();
  ay->SetTitleColor(kBlack);
  ay->SetTitleFont(132);
  ay->SetLabelFont(132);
}



RooPlot * AFitProjectionPlot::makeLRProjectionFrame(RooAbsPdf *pdf, RooDataSet * data, TString varToPlot, TString sigCompName) {
  /*
   * Compute a likelihood ratio projection plot for the ratio ln(Lsig)/ln(total).  
   *
   *    pdf            is the total model 
   *    data           is the data set to projection 
   *    varToPlot      is the discriminating variable to plot
   *    sigCompName    is the signal model name 
   */
  if(data->numEntries() == 0) return 0;
  RooArgSet & aset = *(RooArgSet*)data->get(0);

  /*
   * Using this pdf, now append the likelihood to data
   */
  RooRealVar * var = (RooRealVar*)aset.find(varToPlot);
  RooArgSet * projectionSet = (RooArgSet*)aset.Clone("projectionSet");
  projectionSet->remove(*var);
  projectionSet->Print("v");

  // identify the signal PDF component from the total PDF
  RooArgSet * compSet = (RooArgSet*)pdf->getComponents();
  RooAbsPdf * sigpdf = (RooAbsPdf*)compSet->find(sigCompName);

  // create the pdf projection: one for the total pdf, and one for the signal
  const RooAbsReal * projPDF = pdf->createPlotProjection(*projectionSet, *var);
  const RooAbsReal * projPDFsig = sigpdf->createPlotProjection(*projectionSet, *var);

  // compute the likelihood ratio
  RooFormulaVar llratio("llratio", "log(Lsig/Ltotal)", "log(@0/@1)", RooArgSet(*projPDFsig, *projPDF)); 
  RooRealVar * llVar = (RooRealVar*)data->addColumn(llratio);

  Double_t min, max;
  data->getRange(*llVar, min, max);
  if(min< 0) min = 0;
  cout << "plot range = " << min << " " << max << endl;

  RooPlot * frame = llVar->frame(min, max, 100);
  frame->SetName("llratioFrame");
  frame->SetTitle("");
  data->plotOn(frame);

  return frame;
}

RooPlot * AFitProjectionPlot::makeLRProjection(RooAbsPdf *pdf, RooDataSet * data, TString varToPlot, TString sigCompName, Double_t llrMin) {
  /*
   * Compute a likelihood ratio projection plot for the ratio ln(Lsig)/ln(total).  
   *
   *    pdf            is the total model 
   *    data           is the data set to projection 
   *    varToPlot      is the discriminating variable to plot
   *    sigCompName    is the signal model name 
   *    llrMin         the minimum allowed value of the llratio to plot.  Inspect the projection
   *                   made by the makeLRProjectionFrame fucntion to determine an appropriate 
   *                   value to use.
   */
  if(data->numEntries() == 0) return 0;
  RooArgSet & aset = *(RooArgSet*)data->get(0);

  /*
   * check that llratio is a part of the data set - if not append it
   */
  RooRealVar * llVar = (RooRealVar*)aset.find("llratio");
  if(llVar) {
    cout.width(80);
    cout.fill('_');
    cout << " " << endl;
    cout << "The likelihood ratio variable 'llratio' has been appended to this data set already" << endl;
    cout.width(80);
    cout.fill('_');
    cout << " " << endl;
  } else {
    cout.width(80);
    cout.fill('_');
    cout << " " << endl;
    cout << "The likelihood ratio variable 'llratio' will be appended the data set" << endl;
    cout.width(80);
    cout.fill('_');
    cout << " " << endl;

    RooPlot * llrframe = makeLRProjectionFrame(pdf, data, varToPlot, sigCompName);
  }

  /*
   * make the llratio plot for all data > llrMin, and project the pdf
   * accordingly.
   */
  RooRealVar * var = (RooRealVar*)aset.find(varToPlot);
  TString llrcut = "llratio>";
  llrcut += llrMin;
  llrcut.ReplaceAll(" ", "");
  RooDataSet * sliceData = (RooDataSet*)data->reduce(llrcut);
  cout << "The data slice contains " << sliceData->numEntries() << " events" << endl;

  RooPlot * theframe = var->frame();
  TString framename = var->GetName();
  framename += "_projection";
  theframe->SetName(framename);
  theframe->SetTitle("");
  sliceData->plotOn(theframe);
  pdf->plotOn(theframe, ProjWData(*sliceData),  LineColor(kBlue) );
  pdf->plotOn(theframe, ProjWData(*sliceData),  LineColor(kBlack),  LineStyle(kDashed), Components(sigCompName) );

  return theframe;
}
