/*****************************************************************************
 * Project: RooFit                                                           *
 * Package: RooFitModels                                                     *
 *    File: $Id: AFitRelBreitWignerPdf.cc,v 1.1 2006/04/08 21:59:57 bevan Exp $
 * Authors:                                                                  *
 * Katherine George - University of Liverpool                                * 
 * Adrian Bevan QMUL (ported from RptBaseTools to AFit - 2009)               *
 *****************************************************************************/

// -- CLASS DESCRIPTION [PDF] --
// relativistic breit wigner with blatt weisskopf form factor dependence.

#include <iostream>
#include <math.h>

#include "AFit/AFitRelBreitWignerPdf.hh"
#ifdef __BABARROOFIT__
#include "RooFitCore/RooAbsReal.hh"
#include "RooFitCore/RooRealVar.hh"
#else
#include "RooAbsReal.h"
#include "RooRealVar.h"
#endif

ClassImp(AFitRelBreitWignerPdf)

AFitRelBreitWignerPdf::AFitRelBreitWignerPdf(const char *name, const char *title,
			 RooAbsReal& _x, RooAbsReal& _mean,
			 RooAbsReal& _width, RooAbsReal& _spin, RooAbsReal& _radius,
                         RooAbsReal& _massa, RooAbsReal& _massb) :
  RooAbsPdf(name,title),
  x("x","Dependent",this,_x),
  mean("mean","Mean",this,_mean),
  width("width","Width",this,_width),
  spin("spin", "Spin", this, _spin),
  radius("radius", "radius", this, _radius),
  massa("massa", "mass of daughter A", this, _massa),
  massb("massb", "mass of daughter B", this, _massb)
{
}


AFitRelBreitWignerPdf::AFitRelBreitWignerPdf(const AFitRelBreitWignerPdf& other, const char* name) : 
  RooAbsPdf(other,name), x("x",this,other.x), mean("mean",this,other.mean),
  width("width",this,other.width),spin("spin", this, other.spin), radius("radius", this, other.radius), 
  massa("massa", this, other.massa),  massb("massb", this, other.massb) 
{
}


Double_t AFitRelBreitWignerPdf::evaluate() const
{
  Double_t arg= x*x - mean*mean;  
  Double_t gammaf = Gamma();

  if(x < massa+massb) return 0;
  else return x*x / (arg*arg + x*x*gammaf*gammaf);
}

Double_t AFitRelBreitWignerPdf::Gamma() const
{
  Double_t kx = KFunction((double)x);
  Double_t km = KFunction(mean);
  Double_t fx = FFunction(radius*kx);
  Double_t fm = FFunction(radius*km);
  Double_t rk = 0.0;
  Double_t rf = 0.0;
  if(km!=0){
    rk = (kx/km);
    if(spin ==1){
      rk = rk*rk*rk;
    }
    if(spin ==2){
      rk = rk*rk*rk*rk*rk;
    }    
  }
  if(fm!=0){
    rf = (fx/fm);
  }
  return width*(mean/x)*rk*rf;
}

Double_t AFitRelBreitWignerPdf::FFunction(Double_t X) const
{
  if(spin==0){
      return 1.0;
  }
  if(spin==1){
    return 1.0/(1 + X*X);
  }
  if(spin==2){
    return 1.0/(9 + 3*X*X + X*X*X*X); 
  }
  return 1.0;
}

Double_t AFitRelBreitWignerPdf::KFunction(Double_t X) const
{
  if(X==0) return 0;
  Double_t massone = sqrt(1 - (massa + massb)*(massa + massb)/(X*X));
  Double_t masstwo = sqrt(1 - (massa - massb)*(massa - massb)/(X*X));
  return X/2.0 * massone * masstwo;
}


/*
  Int_t AFitRelBreitWignerPdf::getAnalyticalIntegral(RooArgSet& allVars, RooArgSet& analVars, const char* rangeName) const 
  {
  if (matchArgs(allVars,analVars,x)) return 1 ;
  return 0 ;
  }
  
  
  Double_t AFitRelBreitWignerPdf::analyticalIntegral(Int_t code, const char* rangeName) const 
  {
  switch(code) {
  case 1: 
  {
  Double_t c = 2./width;
  return c*(atan(c*(x.max(rangeName)-mean)) - atan(c*(x.min(rangeName)-mean)));
  }
  }
  
  assert(0) ;
  return 0 ;
  }
*/

