/*****************************************************************************
 * Project: BaBar detector at the SLAC PEP-II B-factory
 * Package: AFit
 *    File: $Id: AFitToy.cc,v 1.5 2005/07/03 23:04:22 bevan Exp $
 * Authors:
 *    A Bevan (a.j.bevan@qmul.ac.uk)
 *
 * Copyright (C) 2008 QMUL
 *****************************************************************************/

// -- CLASS DESCRIPTION [AFit] --
// Utility interface to make toy samples and run toy experiments

#include "AFit/AFitToy.hh"
#include "AFit/AFitMaster.hh"

#ifdef __BABARROOFIT__
#include "RooFitCore/RooRandom.hh"
#include "RooFitCore/RooGlobalFunc.hh"
#include "RooFitCore/RooRealVar.hh"
#else
#include "RooRandom.h"
#include "RooGlobalFunc.h"
#include "RooRealVar.h"
#endif

#include "TIterator.h"
#include "TFile.h"
#include "TTree.h"

using namespace RooFit;

ClassImp(AFitToy)
  ;

AFitToy::AFitToy() 
{
  setSeed();
  setNToys();
  setOutputDir();
  setPoisson();
}

void AFitToy::setSeed(Int_t iSeed){
  _iSeed = iSeed; 

  RooRandom::randomGenerator()->SetSeed( _iSeed );
  gRandom->SetSeed( _iSeed );
}

void AFitToy::dumpFitResult(TString file, RooFitResult * r)
{
/* After running a fit to a toy data sample, the fit result can be converted into
 * a TTree for chaining togeter with other results.  This function facilitates
 * this process.  The fit parameters allowed to float when the 'r' was created
 * are stored in a tree, along with the fit quality parameters.
 * 
 *   o All floated parameters are filled in the ntuple 
 *   o the fit status is filled in the ntuple
 *   o -lnL
 *   o edm
 *   o numInvalidNll
 *   o covQual 
 * 
 *  KEY to fit Status
 * -------------------
 *  status = 0 ---> fit converged and error matrix is accurate
 *         = 4 ---> fit converged and error matrix is uncertain
 * 
 *  KEY to covQual
 * ---------------
 *  covQual = 0 ---> Not calculated at all
 *  covQual = 0 ---> Approximation only, not accurate
 *  covQual = 0 ---> Full matrix, but forced positive-definite
 *  covQual = 0 ---> Full, accurate covariance matrix
 */
  if(!r) return;
  TFile f(file, "RECREATE");

  // make sure one can get the correlation matrix back by storing 
  // the fit result with a known name
  r->SetName("fitResult");
  r->Write();
  TTree  * tree = new TTree("resultdata", "floatParsFinal list of results");
  const RooArgList & pars = r->floatParsFinal();
  RooRealVar * var = 0;
  TIterator * iter = pars.createIterator();
  const Int_t N = pars.getSize();
  Double_t vars[N];
  Double_t evars[N];
  Double_t varsGlobalCorrelation[N];
  Int_t i = 0;
  cout << "appending branches to tree:" << endl;
  while(var = (RooRealVar*)iter->Next()) {
    vars[i] = var->getVal();
    evars[i] = var->getError();
    varsGlobalCorrelation[i] = r->globalCorr(var->GetName());
    
    TString name = var->GetName();
    name += "/D";
    tree->Branch(var->GetName(), &(vars[i]), name.Data());
    
    name = var->GetName();
    name += "Error/D";
    tree->Branch((var->GetName()+TString("Error")).Data(), &(evars[i]), name.Data());
    cout << "\t" << var->GetName() << " + (Error)" << endl;

    name = var->GetName();
    name += "GlobalCorr/D";
    tree->Branch((var->GetName()+TString("GlobalCorr")).Data(), &(varsGlobalCorrelation[i]), name.Data());

    i++;
  }
   
  // add in the fit info - edm, status, nll
  cout << "\tedm, status, minNll, numInvalidNLL" << endl;
  Double_t edm = r->edm();
  Double_t nll = r->minNll();
  Int_t status = r->status();
  Int_t badnll = r->numInvalidNLL();
  Int_t covQual = r->covQual(); 

  tree->Branch("edm",           &edm , "edm/D");
  tree->Branch("covQual",       &covQual , "covQual/I");
  tree->Branch("status",        &status, "status/I" );
  tree->Branch("minNll",        &nll, "minNll/D" );
  tree->Branch("numInvalidNLL", &badnll , "numInvalidNLL/I");

  tree->Fill();

  f.Write();
  f.Close();
}

void AFitToy::chainResults(TChain & chain, Int_t imin, Int_t imax, Bool_t kCutOnStatus) {
  for(Int_t i = imin; i <= imax; i++) {
    TString filename = _odir;
    filename += i;
    filename += "_toyresults.root";
    filename.ReplaceAll(" ", "");

    ifstream in(filename);
    if(!in) {
      cout << "\tUnable to open file " << filename << endl;
    }
    else {
      TFile fi(filename.Data());
      if(fi.IsZombie()) {
	cout << " skipping zombie file: " << filename.Data() << endl;
	continue;
      }
      // check fit status
      if(kCutOnStatus)	{
	TTree * tree = (TTree*)fi.Get("resultdata");
	Int_t status = 0;
	tree->SetBranchAddress("status", &status);
	tree->GetEntry(0);
	if(status == 0) {
	  chain.Add(filename.Data());
	  cout << " adding file: " << filename.Data() << endl;
	}
      } 
      else 
      {
	chain.Add(filename.Data());
	cout << " adding file: " << filename.Data() << endl;
      }
    }
  }
}

void AFitToy::generateToys(TString configFile, TString opt, Int_t iMin, Int_t iMax, RooDataSet * proto){
  for(int i=iMin; i <= iMax ; i++) {
    cout << " __________________________________________________________________________________"<<endl;
    cout <<  "AFitToy::generateToys(...)" << endl;
    cout << " Generating toy number " << i << " from config file " << configFile << endl;
    cout << " __________________________________________________________________________________"<<endl;
    generateToySample(configFile, opt, i, proto);
  } 
}

RooDataSet * AFitToy::generateToySample(TString configFile, TString opt, Int_t iToy, RooDataSet * proto) {
  /*
   * Generate a toy experiment and run fit to the data
   *
   *   f = fit using the AFitMaster
   *   p = persist data
   *   r = persist results
   *
   * If opt contains: 
   *      'f' the generated data sample will be fitted
   *
   *      'p' a root file will be written to the default output directory
   *          with a name of <iToy>_toydata.root
   *
   *      'r' a root file will be written to the default output directory
   *          with a name of <iToy>_toyresults.root
   *
   */
  AFitMaster master(configFile);
  RooAbsPdf * pdf = master.getPdf();
  Bool_t fitData           = kFALSE;
  Bool_t persistToyData    = kFALSE;
  Bool_t persistToyResults = kFALSE;

  opt.ToLower();
  if(opt.Contains("f")) fitData = kTRUE;
  if(opt.Contains("p")) persistToyData = kTRUE;
  if(opt.Contains("r")) persistToyResults = kTRUE;

  cout << "\tAFitToy::Generating a data set for toy with seed " << iToy << endl;
  setSeed(iToy);
  RooDataSet * data = generateToySample(configFile, proto);
  data->SetName("data");

  RooFitResult * r(0);
  if(fitData) {
    cout << "\tAFitToy::Fitting generated toy data sample" << endl;
    /*
     * fit the data, ensuring that a fit result is obtained from the fit
     */
    RooStringVar & fitOpt = master.getFitOpt();
    TString options = fitOpt.getVal();
    if(!options.Contains("r")) options += "r";
    r = pdf->fitTo(*(RooAbsData*)data, FitOptions(options));
  }

  if(persistToyData) {
    /*
     * Save the data to a root file for reuse
     */
    TString opfilename = _odir;
    opfilename += _iSeed;
    opfilename += "_toydata.root";
    opfilename.ReplaceAll(" ", "");
    cout << "\tAFitToy::Persisting toy data file " << opfilename << endl;
    TFile of(opfilename.Data(), "RECREATE");
    of.cd();
    data->Write();
    of.Write();
    of.Close();
  }

  if(persistToyResults && r) {
    /*
     * Store the fit result, and a tree of the fit result in a root file
     * for analysis.
     */  
    TString opfilename = _odir;
    opfilename += _iSeed;
    opfilename += "_toyresults.root";
    opfilename.ReplaceAll(" ", "");
    cout << "AFitToy::Persisting toy results in " << opfilename << endl;
    dumpFitResult(opfilename, r);
  }

  return data;
}

RooDataSet * AFitToy::generateToySample(TString configFile, RooDataSet * proto) {
  /*
   * Generate a toy sample to fit based on a specific configuration file 
   * If a prototype is required for conditional variables, this should be 
   * passed to this function. 
   *
   * The number of events to generate is the sum of events specified 
   * for the components listed in the configuration file.
   */
  cout << " have set the configuration file to \'" << configFile << "\'" << endl;
  AFitMaster master(configFile);
  RooAbsPdf * pdf = master.getPdf();

  if(!pdf) {
    cout << "AFitToy" << "::generateToySample(...) ERROR - failed to get PDF from AFitMaster!" << endl;
    return 0;
  } else {
    cout << " Have succesfully made the PDF: " << pdf << endl;
  }
  RooArgSet * discVars = &master.getDiscVarSet();
  RooArgSet * coefVars = &master.getCoefSet();
  cout << "--------------- Discriminating Variable List --------------" << discVars << endl;
  discVars->Print();
  cout << "--------------- Coefficient Variable List --------------" << coefVars << endl;
  coefVars->Print();


  /*
   * Determine number of events to generate from configuration file
   */
  Double_t nToGen(0);
  TIterator * iter = coefVars->createIterator();
  TObject * obj;
  while(obj = iter->Next()) {
    RooRealVar &thisCoefVar = *(RooRealVar*)obj;
    nToGen += thisCoefVar.getVal();
    cout << " Coefficient: " << thisCoefVar.GetName() << " value = " << thisCoefVar.getVal() << endl; 
  }

  cout << "nToGen = " << nToGen << endl;
  /*
   * Determine variables to generate, making sure that these are not duplicated by proto
   */
  RooDataSet * data(0);
  if(proto) {
    const RooArgSet * protoSet = proto->get(0);
    RooArgSet varsToGen;
    TIterator * iter = discVars->createIterator();
    TObject * obj;
    while(obj = iter->Next()) {
      RooRealVar &thisDiscVar = *(RooRealVar*)obj;
      if(!protoSet->find( thisDiscVar.GetName() )) {
	varsToGen.add(thisDiscVar);
      }
    }

    cout << "AFitToy" << "::generateToySample(...) - have selected the following variables to generate" << endl;
    varsToGen.Print("v");
    cout << "AFitToy" << "::generateToySample(...) will use the following prototype data set in generation" << endl;
    proto->Print();
    data = generateToySample(pdf, varsToGen, (Int_t)nToGen, proto);
  } else {
    RooArgSet myDiscVars(*discVars);
    data = generateToySample(pdf, myDiscVars, (Int_t)nToGen, 0);
  }

  return data;
}

RooDataSet * AFitToy::generateToySample(RooAbsPdf * pdf, RooArgSet & varsToGen, Int_t nToGen, RooDataSet * proto) {
  /*
   * Given a PDF, list of variables to generate, number of 
   * events to generate and prototype data set, generate a 
   * simulated set of nToGen events.
   *
   * pdf       = PDF to generate from
   * varsToGen = variables to generate (not including variables in prototype data set
   * nToGen    = number of events to generate (if _iPois is true, then a random number 
   *             about this poisson mean is used.
   * proto     = prototype data set for use in generating any conditional variables.
   *             The prototype may be null.
   */
  RooDataSet * data(0);
  if(pdf && nToGen > 0) {
    Int_t iNToGenerate = nToGen;
    if(_iPois) {
      iNToGenerate = (int)( gRandom->Poisson( nToGen ) );
    }
    if(proto) data = (RooDataSet*)pdf->generate(varsToGen, *proto, iNToGenerate);
    else data = (RooDataSet*)pdf->generate(varsToGen, iNToGenerate);
    Int_t nGen = data->numEntries();
    if(nGen != iNToGenerate) {
      cout << "AFitToy" << "::generateToySample(...) ERROR generated " << nGen << " out of " 
	   << iNToGenerate << " events for " << pdf->GetName() ;
      if(proto) cout << " (proto-type has " << proto->numEntries() << " events)";
      cout << endl;
    }
  }
  return data;
} 

void AFitToy::generateEmbeddedToys(RooDataSet * data, Int_t nToGet, Int_t iMin, Int_t iMax, TString cut, TString type) {
  /*
   * Generate samples containing nToGet entries with _iSeed = iMin, through iMax.  The supplied data sample
   * is used as the source for sampling, and any cuts to be applied to the data are done prior to 
   * entering the loop over _iSeed.
   */
  if(!data) return;
  RooDataSet * reduceddata = data;
  if(cut!="1") {
    reduceddata = (RooDataSet*)data->reduce(cut);
  }

  for(int i=iMin; i <= iMax ; i++) {
    _iSeed = i;
    cout << " __________________________________________________________________________________"<<endl;
    cout <<  "AFitToy::generateEmbeddedToys(...)" << endl;
    cout << " Generating toy number " << i << " from dataset " << data->GetName() << endl;
    cout << " __________________________________________________________________________________"<<endl;
    generateEmbeddedToySample(reduceddata, i, nToGet, type);
  } 
}

void AFitToy::generateEmbeddedToySample(RooDataSet * data, Int_t iToy, Int_t nToGet, TString type) {
  /*
   * Obtain a sub-sample of data for use in an embedded toy, and write out the 
   * sampled data file to _odir.
   */

  RooDataSet * thisData = generateEmbeddedToySample(data, nToGet);
  thisData->SetName("data");
  thisData->SetTitle("data sample for embedded toy study");
  TString opfilename = _odir;
  opfilename += _iSeed;
  opfilename += "_embtoydata";
  opfilename += type;
  opfilename += ".root";
  opfilename.ReplaceAll(" ", "");
  cout << "\tAFitToy::Persisting toy data file " << opfilename << endl;
  TFile of(opfilename.Data(), "RECREATE");
  of.cd();
  thisData->Write();
  of.Write();
  of.Close();
}

RooDataSet * AFitToy::generateEmbeddedToySample(RooDataSet * data, Int_t nToGet, TString cut) {
  /*
   * Generate a random sample of events to use in an embedded toy MC.  If necessary, cut on the data sample
   * prior to sampling.
   */
  if(data->numEntries() < nToGet) return 0;

  RooDataSet * theData(0);
  if(cut == "1") theData = data;
  else theData = (RooDataSet*)data->reduce(cut);

  RooDataSet *thisSample = generatePrototype(theData, nToGet, kFALSE);

  return thisSample;
}

RooDataSet * AFitToy::generatePrototype(RooDataSet *data, Int_t nToGen, Bool_t returnAnyway)
{
  /*
   * get a random sub-sample of an input prototype data set for 
   * generation.  
   * 
   * If the number of events in the proto-type source data set, 
   * then this function will return a null data set unless 
   * returnAnyway is true.  The default is to return a null
   * data set.
   */

  if(!data) return 0;
  if(nToGen > data->numEntries())    {
    if(returnAnyway) return data;
    cout << "AFitToy" << "::generatePrototype(..) ERROR: Requested more events that exist in source data - returning a null data set" << endl;
    return 0;
  }
  const RooArgSet & aset = *(data->get(0));
  RooDataSet * newDataSet = new RooDataSet("prototype", "sampled proto-type data", aset);
  const Int_t nDat = (Int_t)data->numEntries();
  Int_t iused[nDat];
  for(Int_t iDat = 0; iDat < nDat; iDat++) {
    iused[iDat] = 0;
  }
  while(newDataSet->numEntries() < nToGen) {
    float R=gRandom->Rndm();
    
    Int_t thisI = (Int_t)(nDat * R);
    if(!iused[thisI]) {
      iused[thisI] = 1;
      const RooArgSet * aset = data->get(thisI);
      newDataSet->add(*aset);
    }
  }
  Int_t nGen = newDataSet->numEntries(); 
  if( nGen != nToGen ) {
      cout << "AFitToy" << "::generatePrototype(...) ERROR generated " << nGen << " out of " 
	   << nToGen << " events for this proto-type" << endl;
  }
  return newDataSet;
}

void AFitToy::mergeFiles(TObjArray * farray, TString opfile, TString dataname) {
  /*
   * Open a list of files, and append the RooDataSets in each file so that 
   * a single output file can be written that contains all of the merged data.
   */
  RooDataSet * data(0);  
  for(Int_t i=0; i < farray->GetEntries(); i++) {
    TString thisFile = ((TObjString*)farray->At(i))->GetString();
    TFile * f = new TFile(thisFile);
    RooDataSet * thisData = (RooDataSet *)f->Get(dataname);
    if(i==0) data = (RooDataSet*)thisData->Clone();
    else data->append(*(RooDataSet*)thisData->Clone());
    f->Close();
  }

  data->Print("v");
  TFile fo(opfile, "RECREATE");
  fo.cd();
  data->Write();
  fo.Write();
  fo.Close();
}
